############################################################################
# tools/pynuttx/nxgdb/circbuf.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
from typing import Generator, Union

import gdb

from . import autocompeletion, utils
from .protocols import circbuf as p


class CircBuf(utils.Value, p.CircBuf):
    def __init__(
        self, obj: Union[gdb.Value, utils.Value], datatype: gdb.Type = None
    ) -> None:
        circbuf_s = utils.lookup_type("struct circbuf_s")
        if obj.type.code == gdb.TYPE_CODE_INT:
            obj = obj.cast(circbuf_s.pointer())

        if obj.type.code == gdb.TYPE_CODE_PTR:
            obj.cast(circbuf_s.pointer())
            obj = obj.dereference()

        super().__init__(obj)

        # datatype must not be a pointer, because we are going to construct value from memory
        if not datatype:
            datatype = utils.lookup_type("char")

        if isinstance(datatype, str):
            datatype = utils.lookup_type(datatype)

        if datatype.code == gdb.TYPE_CODE_PTR:
            datatype = datatype.target()

        self.datatype = datatype

    def __str__(self) -> str:
        return (
            f"(struct circbuf_s *){hex(self.address)} base: {self.base} "
            f"size: {self.size} head: {self.head} tail: {self.tail}"
        )

    @property
    def size(self) -> int:
        return int(self["size"])

    @property
    def used(self) -> int:
        return int(self["head"]) - int(self["tail"])

    @property
    def space(self) -> int:
        return self.size - self.used

    @property
    def is_inited(self) -> bool:
        return bool(self["base"])

    @property
    def is_empty(self) -> bool:
        return self.used == 0

    @property
    def is_full(self) -> bool:
        return not self.space

    def _peekat(self, pos, len) -> memoryview:
        if len > self.size:
            return None

        pos = pos % self.size
        total = len
        if pos + len > self.size:
            len = self.size - pos

        memory = gdb.selected_inferior().read_memory(self.base + pos, len)
        if len < total:
            memory += gdb.selected_inferior().read_memory(self.base, total - len)
        return memory

    @property
    def history(self) -> Generator[utils.Value, None, None]:
        """Iterate over the history data in the circbuf_s, from oldest to newest"""
        if not self.base or not self.size:
            # Uninitialized buffer
            return []

        head = int(self.head)
        size = int(self.size)
        sizeof = self.datatype.sizeof

        if head < size:
            # The buffer is never wrapped, read from the beginning
            offset = 0
            end = head
        else:
            # The buffer is wrapped, read from the head
            offset = head % size
            end = offset + size

        while offset < end:
            memory = self._peekat(offset, sizeof)
            value = gdb.Value(memory, self.datatype)
            yield value
            offset += sizeof

    @property
    def unread(self) -> Generator[utils.Value, None, None]:
        """Return all unread data in circle buffer"""
        if not self.base or not self.size:
            return []

        # Read from tail towards head for all data.
        tail = int(self.tail)
        head = int(self.head)
        sizeof = self.datatype.sizeof
        offset = tail
        while offset < head:
            memory = self._peekat(offset, sizeof)
            yield gdb.Value(memory, self.datatype)
            offset += sizeof


@autocompeletion.complete
class CircBufInfo(gdb.Command):
    """Print circbuf_s information"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description="Dump circle buffer information")
        parser.add_argument(
            "--type",
            type=str,
            help="The data type the circbuf_s contains",
            default=None,
        )
        parser.add_argument(
            "--history",
            action="store_true",
            help="Dump the history data in the circbuf_s",
        )
        parser.add_argument(
            "--unread",
            action="store_true",
            help="Dump the unread data in the circbuf_s",
        )
        parser.add_argument(
            "address",
            type=str,
            help="The address of the circubuf_s",
        )
        return parser

    def __init__(self):
        super().__init__("circbuf", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        try:
            args = self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            gdb.write("Invalid arguments\n")
            return

        entry = utils.parse_and_eval(args.address)
        circbuf = CircBuf(entry, datatype=args.type)

        print(circbuf)  # Dump buffer basic information

        if args.history:
            dumpdata = circbuf.history
        elif args.unread:
            dumpdata = circbuf.unread
        else:
            dumpdata = []

        print(f"Dumping data with type {args.type}")
        for i, data in enumerate(dumpdata):
            print(f"{i}: {data.format_string(styling=True)}")



#!/usr/bin/python3
############################################################################
# tools/pynuttx/gdbrpc/examples/memleak.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
import os
import queue
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))

import gdbrpc  # noqa: E402


class FetchResult(gdbrpc.PostRequest):
    def __init__(self):
        super().__init__()

    def __call__(self, result):
        print(result)


class MemLeak(gdbrpc.Request):
    def __init__(self):
        super().__init__()

    def __call__(self, q: queue.Queue):
        import gdb

        q.put(gdb.execute("memleak", to_string=True))


if __name__ == "__main__":

    client = gdbrpc.Client(logLevel=logging.DEBUG)

    assert client.connect()

    bt = FetchResult()
    client.call(MemLeak(), bt)

    print("Waiting for async call to complete...")

    bt.finish.wait()

    print("Async call completed.")

    client.disconnect()



############################################################################
# tools/pynuttx/nxgdbmcp/src/gmcp/utils.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import socket

from mcp.server.fastmcp import Context

from .context import get_session


def error_handler(func):
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except ValueError as e:
            return str(e)
        except Exception as e:
            return f"Error occurred: {str(e)}"

    return wrapper


@error_handler
async def _exec_command(ctx: Context, session_id: str, command: str) -> str:
    session = get_session(ctx, session_id)
    output = await session.execute_command(command)
    return f"Executed command: {command}\n\nOutput:\n{output}"


def get_available_port() -> int:
    """Find an available TCP port on localhost."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("localhost", 0))
        return s.getsockname()[1]



############################################################################
# tools/pynuttx/gdbrpc/examples/gdbscript.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import os
import queue

import gdbrpc


class ScriptLoader(gdbrpc.Request):
    def __init__(self, path):
        super().__init__()
        self.path = path

    def __call__(self, q: queue.Queue):
        # Executed on GDB, the server side
        import gdb

        try:
            result = gdb.execute(f"source {self.path}", to_string=True)
        except Exception as e:
            result = str(e)
        q.put(result)


if __name__ == "__main__":

    client = gdbrpc.Client()  # Use default localhost:20819, change if needed

    assert client.connect()

    script_path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "examplescript.py")
    )

    print(client.call(ScriptLoader(script_path), timeout=10))

    client.disconnect()



############################################################################
# tools/pynuttx/nxgdb/stack.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import traceback

import gdb

from . import utils

STACK_COLORATION_PATTERN = 0xDEADBEEF
CONFIG_STACK_COLORATION = utils.get_global_symbol("up_check_tcbstack") is not None


class Stack(object):
    def __init__(self, base, size, cursp):
        # We don't care about the stack growth here, base always point to the lower address!
        self._stack_base = base
        self._stack_top = base + size
        self._cur_sp = cursp
        self._stack_size = size

        self._sanity_check()

    def _sanity_check(self):
        # do some basic sanity checking to make sure we have a sane stack object
        if (
            self._stack_base > self._stack_top
            or not self._stack_size
            or self.is_stackof()
        ):

            gdb.write(
                f"base: {hex(self._stack_base)}, \
                size: {hex(self._stack_size)}, sp: {hex(self._cur_sp)}\n"
            )

            gdb.write("Inconsistant stack size...Maybe memory corruption?\n")
            gdb.write("dumping the stack:\n")

            ptr_4bytes = gdb.Value(self._stack_base).cast(
                utils.lookup_type("unsigned int").pointer()
            )

            for i in range(0, self._stack_size // 4):
                if i % 8 == 0:
                    gdb.write(f"{hex(self._stack_base + 4 * i)}: ")

                gdb.write(f"{hex(ptr_4bytes[i]):10} ")

                if i % 8 == 7:
                    gdb.write("\n")

            gdb.write("\n")
            gdb.GdbError(
                "pls check your stack size! sp:{0:x} base:{1:x}".format(
                    self._cur_sp, self._stack_base
                )
            )

    def cur_usage(self):
        return self._stack_top - self._cur_sp

    def check_stack_usage(self) -> int:
        np = utils.import_check("numpy", errmsg="Please pip install numpy\n")
        if not np:
            raise gdb.GdbError(
                "stack max usage check requires numpy, please install it via pip"
            )
        memory = gdb.selected_inferior().read_memory(self._stack_base, self._stack_size)

        color_size = utils.sizeof("int")
        pattern = int(STACK_COLORATION_PATTERN).to_bytes(color_size, byteorder="little")

        arr = np.frombuffer(memory, dtype=np.uint8)
        arr = arr.reshape(-1, color_size)
        pattern_arr = np.frombuffer(pattern * (len(arr)), dtype=np.uint8).reshape(
            -1, color_size
        )

        mismatch = np.any(arr != pattern_arr, axis=1)
        first_diff = np.argmax(mismatch)

        if not mismatch.any():
            used = 0
        else:
            used = self._stack_size - (first_diff * color_size)

        return used

    def max_usage(self):
        if not CONFIG_STACK_COLORATION:
            return 0
        return self.check_stack_usage()

    def avalaible(self):
        return self._stack_size - self.cur_usage()

    def maxdepth_backtrace(self):
        raise gdb.GdbError("Not implemented yet", traceback.print_stack())

    def is_stackof(self) -> bool:
        # we should notify the user if the stack overflow is about to happen as well!
        if utils.check_inferior_valid():
            return (self.check_stack_usage() / self._stack_size) >= 95
        return False


# Always refetch the stack infos, never cached as we may have threads created/destroyed
# dynamically!
def fetch_stacks():
    stacks = dict()

    for tcb in utils.get_tcbs():
        # We have no way to detect if we are in an interrupt context for now.
        # Originally we use `and not utils.in_interrupt_context()`
        sp = utils.get_sp(tcb)

        try:
            stacks[int(tcb["pid"])] = Stack(
                int(tcb["stack_base_ptr"]),
                int(tcb["adj_stack_size"]),
                sp,
            )

        except gdb.GdbError as e:
            pid = tcb["pid"]
            gdb.write(
                f"Failed to construction stack object for tcb {pid} due to: {e}\n"
            )

    return stacks


class StackUsage(gdb.Command):
    """Display the stack usage of each thread, similar to cat /proc/<pid>/stack"""

    def __init__(self):
        super().__init__("stack-usage", gdb.COMMAND_USER)
        self._stacks = []
        # format template
        self._fmt = (
            "{0: <4} | {1: <10} | {2: <10} | {3: <20} | {4: <10} | {5: <15} | {6: <15}"
        )

    def format_print(self, pid, stack, tcb):
        def gen_usage_str(x):
            usage = x / stack._stack_size
            res = f"{usage:.2%}"
            if usage > 0.8:
                res += "!"
            return res

        def gen_info_str(x):
            res = f"{str(x)} -> {gen_usage_str(x)}"
            return res

        if hasattr(self, "table"):
            self.table.add_row(
                [
                    pid,
                    utils.get_task_name(tcb),
                    hex(tcb["entry"]["pthread"]),
                    hex(stack._stack_base),
                    stack._stack_size,
                    str(stack.cur_usage()),
                    gen_usage_str(stack.cur_usage()),
                    str(stack.max_usage()),
                    gen_usage_str(stack.max_usage()),
                ]
            )
        else:
            gdb.write(
                self._fmt.format(
                    pid,
                    utils.get_task_name(tcb),
                    hex(tcb["entry"]["pthread"]),
                    hex(stack._stack_base),
                    stack._stack_size,
                    gen_info_str(stack.cur_usage()),
                    gen_info_str(stack.max_usage()),
                )
            )
            gdb.write("\n")

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        stacks = fetch_stacks()

        args = [int(arg) for arg in args.split()]

        pids = stacks.keys() if len(args) == 0 else args

        prettytable = utils.import_check(
            "prettytable",
            errmsg="Execute `pip install prettytable` for better printing result.\n",
        )
        if prettytable:
            self.table = prettytable.PrettyTable()
            self.table.align = "l"
            self.table.field_names = [
                "PID",
                "NAME",
                "Entry",
                "Base",
                "Size",
                "CurUsage",
                "CurUsage%",
                "MaxUsage",
                "MaxUsage%",
            ]
        else:
            gdb.write(
                self._fmt.format(
                    "Pid", "Name", "Entry", "Base", "Size", "CurUsage", "MaxUsage"
                )
            )
            gdb.write("\n")

        for pid in pids:
            stack = stacks.get(pid)

            if not stack:
                continue

            self.format_print(pid, stack, utils.get_tcb(pid))
        if hasattr(self, "table"):
            gdb.write(f"{self.table.get_string()}\n")

    def diagnose(self, *args, **kwargs):
        return {
            "title": "Stack Usage Report",
            "summary": "Stack usage report",
            "command": "stack-usage",
            "result": "info",
            "category": utils.DiagnoseCategory.memory,
            "message": gdb.execute("stack-usage", to_string=True),
        }



############################################################################
# tools/pynuttx/nxstub/proxy.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import atexit
import logging
import queue
import socket
import threading
from binascii import unhexlify

from . import utils
from .target import Target


class TargetProxy(Target):
    def __init__(self, elf, arch=None, port=None):
        super().__init__(elf, arch=arch)
        self.exit = False
        self.logger = logging.getLogger(__name__)

        def send_thread(queue: queue.Queue, sock: socket.socket):
            while not self.exit:
                packet = queue.get()
                self.logger.debug(f"Send: {packet}")
                sock.send(packet)
                self.logger.debug("Sent.")

        def recv_thread(queue: queue.Queue, sock: socket.socket):
            while not self.exit:
                packet = utils.get_packet(sock)
                queue.put(packet)
                self.logger.debug(f"Received: {packet}")

        sock = self.connect_target(port)
        self.txqueue = queue.Queue()
        self.rxqueue = queue.Queue()
        self.sender = threading.Thread(target=send_thread, args=(self.txqueue, sock))
        self.receiver = threading.Thread(target=recv_thread, args=(self.rxqueue, sock))

        self.sender.start()
        self.receiver.start()

        self.clear_rxqueue()
        atexit.register(self.__del__)

    def __del__(self):
        self.logger.warning("Destructor called")
        self.exit = True
        self.txqueue.put(None)
        self.rxqueue.put(None)
        self.sender.join()
        self.receiver.join()

    def connect_target(self, port):
        target = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        target.connect(("localhost", port))
        self.logger.info(f"Connected to target: localhost:{port}")
        return target

    def clear_rxqueue(self):
        while (packet := self.receive_packet(block=False)) is not None:
            self.logger.debug(f"Clearing queue: {packet}")

    def send_packet(self, packet, timeout=None, noresponse=False):
        encoded = utils.encode_packet(packet)
        checksum = sum(encoded) & 0xFF
        encoded = b"$" + encoded + b"#" + b"%02x" % checksum

        self.clear_rxqueue()
        self.txqueue.put(encoded)
        ack = self.receive_packet(timeout)
        if ack != b"+":
            self.logger.warning(f"Not ack received: {ack}, try again...")
            ack = self.receive_packet(timeout=1)
            if ack != b"+":
                self.logger.error(f"Failed to send packet: {packet}, get ack: '{ack}'")
                return

        self.logger.debug(f"Sent packet: {packet}")
        if noresponse:
            return

        response = self.receive_packet(timeout)
        self.logger.debug(f"Target response: {response}")
        return response

    def receive_packet(self, timeout=1, block=True):
        try:
            return self.rxqueue.get(block=block, timeout=timeout)
        except queue.Empty:
            return None

    def memory_read(self, address: int, length: int) -> bytes:
        self.logger.debug(f"Read target: {address:#x} {length}Bytes")
        response = self.send_packet(b"m%x,%x" % (address, length))
        data = unhexlify(response)
        return data

    def memory_write(self, address, data):
        # We do not support write for now for debug purpose
        self.logger.warning(f"Ignore write target: {address:#x} {len(data)}Bytes")
        pass

    def read_registers(self):
        response = self.send_packet(b"g")
        response = unhexlify(response)
        self.logger.debug(f"Raw registers: {response}")
        return response

    def update_threads(self):
        threads = super().update_threads()
        # Since we are doing live debugging, we need to correct the registers of running threads
        for t in threads:
            if t.state == "Running":
                t.registers.from_g(self.read_registers())
                break
        return threads

    def cont(self, callback):
        self.send_packet(b"c", noresponse=True)  # continue packet has no response

        self.running = True

        def checker(callback):
            self.logger.debug("Cont checker started, waiting for response...")
            while self.running:
                self.logger.debug("Checking...")
                if response := self.receive_packet(timeout=0.1):
                    self.logger.debug(f"Cont got response: {response}")
                    self.running = False
                    self.send_packet(b"z0")
                    self.update_threads()
                    callback(response)
                    return

            self.logger.error(f"Cont checker stopped, target: {self.state}")

        threading.Thread(target=checker, args=(callback,)).start()

    def stop(self):
        # Send a break signal, instead of a full packet.
        self.txqueue.put(b"\x03")

    def forward_packet(self, packet):
        self.txqueue.put(packet)
        return self.receive_packet()



############################################################################
# tools/pynuttx/nxgdb/utils.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from __future__ import annotations

import hashlib
import importlib
import json
import math
import re
import sys
import traceback
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

import gdb
from nxelf.macros import fetch_macro_info, try_expand

from .protocols.thread import Tcb

g_symbol_cache = {}
g_type_cache = {}
g_macro_ctx = None
g_backtrace_cache = {}
TypeOrStr = Union[gdb.Type, str]


class Value(gdb.Value):
    def __init__(self, obj: Union[gdb.Value, Value]):
        super().__init__(obj)

    def __isabstractmethod__(self):
        # Added to avoid getting error using __getattr__
        return False

    def __getattr__(self, key):
        if hasattr(super(), key):
            value = super().__getattribute__(key)
        else:
            value = super().__getitem__(key)

        return Value(value) if not isinstance(value, Value) else value

    def __getitem__(self, key):
        value = super().__getitem__(key)
        return Value(value) if not isinstance(value, Value) else value

    def __format__(self, format_spec: str) -> str:
        try:
            return super().__format__(format_spec)
        except TypeError:
            # Convert GDB value to python value, and then format it
            type_code_map = {
                gdb.TYPE_CODE_INT: int,
                gdb.TYPE_CODE_PTR: int,
                gdb.TYPE_CODE_ENUM: int,
                gdb.TYPE_CODE_FUNC: hex,
                gdb.TYPE_CODE_BOOL: bool,
                gdb.TYPE_CODE_FLT: float,
                gdb.TYPE_CODE_STRING: str,
                gdb.TYPE_CODE_CHAR: lambda x: chr(int(x)),
            }

            t = self.type
            while t.code == gdb.TYPE_CODE_TYPEDEF:
                t = t.target()

            type_code = t.code
            try:
                converter = type_code_map[type_code]
                return f"{converter(self):{format_spec}}"
            except KeyError:
                raise TypeError(
                    f"Unsupported type: {self.type}, {self.type.code} {self}"
                )

    @property
    def address(self) -> Value:
        value = super().address
        return value and Value(value)

    def cast(self, type: str | gdb.Type, ptr: bool = False) -> Optional["Value"]:
        try:
            gdb_type = lookup_type(type) if isinstance(type, str) else type
            if ptr:
                gdb_type = gdb_type.pointer()
            return Value(super().cast(gdb_type))
        except gdb.error:
            return None

    def dereference(self) -> Value:
        return Value(super().dereference())

    def reference_value(self) -> Value:
        return Value(super().reference_value())

    def referenced_value(self) -> Value:
        return Value(super().referenced_value())

    def rvalue_reference_value(self) -> Value:
        return Value(super().rvalue_reference_value())

    def const_value(self) -> Value:
        return Value(super().const_value())

    def dynamic_cast(self, type: gdb.Type) -> Value:
        return Value(super().dynamic_cast(type))


class Symbol:
    def __init__(self, address: Union[int, gdb.Value]):
        self.func: str = ""
        self.address: int = 0
        self._symtab_and_line: gdb.Symtab_and_line = None
        self.resolve_symbol(address)

    def resolve_symbol(self, address: Union[int, gdb.Value]) -> None:
        """Resolve symbol information for the given address"""
        if not address:
            return

        # Check cache first
        if int(address) in g_backtrace_cache:
            cached = g_backtrace_cache[int(address)]
            self.address, self.func, self._symtab_and_line = cached
            return

        # Convert to proper gdb.Value if needed
        if type(address) is int:
            address = gdb.Value(address)

        # Ensure we have a pointer type
        if address.type.code is not gdb.TYPE_CODE_PTR:
            address = address.cast(gdb.lookup_type("void").pointer())

        # Get symbol information
        self.address = int(address)
        self._symtab_and_line = gdb.find_pc_line(int(address))
        if not self.is_valid():
            return

        self.func = address.format_string(symbols=True, address=False)

        # Cache the result
        result = (self.address, self.func, self._symtab_and_line)
        g_backtrace_cache[self.address] = result

    def is_valid(self) -> bool:
        return (
            self._symtab_and_line
            and self._symtab_and_line.is_valid()
            and self._symtab_and_line.symtab
            and self._symtab_and_line.symtab.is_valid()
        )

    @property
    def funcname(self) -> str:
        return self.func

    @property
    def filename(self) -> str:
        if not self.is_valid():
            return ""
        return str(self._symtab_and_line.symtab.fullname())

    @property
    def line(self) -> int:
        if not self.is_valid():
            return 0
        return int(self._symtab_and_line.line)

    def __repr__(self) -> str:
        return (
            f"<Symbol {hex(self.address)}: {self.func} at {self.filename}:{self.line}>"
        )

    def toJSON(self):
        return {
            "address": self.address,
            "function": self.funcname,
            "source": self.filename,
            "line": self.line,
        }


class DiagnoseCategory(str, Enum):
    sched = "sched"
    memory = "memory"
    system = "system"
    power = "power"
    libuv = "libuv"
    rpc = "rpc"
    fs = "fs"
    sensor = "sensor"
    connectivity = "connectivity"
    graphics = "graphics"
    audio = "audio"
    video = "video"


def lookup_type(name, block=None) -> gdb.Type:
    """Return the type object of a type name"""
    key = (name, block)
    if key not in g_type_cache:
        try:
            g_type_cache[key] = (
                gdb.lookup_type(name, block=block) if block else gdb.lookup_type(name)
            )
        except gdb.error:
            g_type_cache[key] = None

    return g_type_cache[key]


def get_fieldnames(t: TypeOrStr) -> List[str]:
    """Return the field names of a type"""
    if isinstance(t, str):
        t = lookup_type(t)

    return [f.name for f in t.fields()]


def get_type_field(obj: Union[TypeOrStr, gdb.Value], field: str) -> gdb.Field:
    """
    Get the type field descriptor from a type or string, or value object.
    """

    if isinstance(obj, str):
        t = lookup_type(obj)
    elif isinstance(obj, gdb.Type):
        t = obj
    elif isinstance(obj, gdb.Value):
        t = obj.type
    else:
        raise gdb.GdbError(f"Unsupported type {type(obj)}")

    if not t:
        return None

    while t.code in (gdb.TYPE_CODE_PTR, gdb.TYPE_CODE_ARRAY, gdb.TYPE_CODE_TYPEDEF):
        t = t.target()

    for f in t.fields():
        if f.name == field:
            return f

        # Check anonymous(f.name is None) fields of struct and union recursively
        if f.name is None and f.type.code in (
            gdb.TYPE_CODE_STRUCT,
            gdb.TYPE_CODE_UNION,
        ):
            if f := get_type_field(f.type, field):
                return f


def get_field_nitems(t: TypeOrStr, field: str) -> int:
    """Return the array length of a field in type, or 0 if no such field"""
    if field := get_type_field(t, field):
        return nitems(field)

    return 0


def get_static_symbol(name: str, domain=None) -> Optional[gdb.Symbol]:
    """
    Return a global symbol with static linkage by name.

    Looks up all static symbols matching the given name and returns the first
    one that is not optimized out. This is useful when multiple static symbols
    with the same name exist across different compilation units, and some may
    be optimized away by the compiler.

    Args:
        name: The symbol name to look up
        domain: Optional symbol domain (e.g., gdb.SYMBOL_VAR_DOMAIN)

    Returns:
        The first non-optimized static symbol, or None if not found
    """
    try:
        symbols = (
            gdb.lookup_static_symbols(name, domain=domain)
            if domain is not None
            else gdb.lookup_static_symbols(name)
        )

        for sym in symbols:
            if sym.value().is_optimized_out:
                continue
            return sym
    except gdb.error:
        pass

    return None


def get_global_symbol(name: str, domain=None) -> Optional[gdb.Symbol]:
    """Return the global symbol object"""
    return (
        gdb.lookup_global_symbol(name, domain=domain)
        if domain is not None
        else gdb.lookup_global_symbol(name)
    )


def get_static_var(name: str) -> Optional[gdb.Symbol]:
    return get_static_symbol(name, domain=gdb.SYMBOL_VAR_DOMAIN)


def get_global_var(name: str) -> Optional[gdb.Symbol]:
    """Return global symbol by name, including static linkage symbols"""
    return gdb.lookup_global_symbol(
        name, domain=gdb.SYMBOL_VAR_DOMAIN
    ) or get_static_var(name)


def get_global_func(name: str) -> Optional[gdb.Symbol]:
    """Return global function symbol by name"""
    return gdb.lookup_global_symbol(name, domain=gdb.SYMBOL_FUNCTION_DOMAIN)


def get_global_func_block(name: str) -> Optional[gdb.Block]:
    """Return the function block by function name"""
    # name to address
    symbol = get_global_func(name)
    if not symbol:
        return None

    address = symbol.value().address
    # address to block
    return gdb.block_for_pc(int(address))


long_type = lookup_type("long")


def dont_repeat_decorator(func):
    def wrapper(self, args, from_tty):
        try:
            self.dont_repeat()
            func(self, args, from_tty)
        except Exception as e:
            print(f"Error: {e}\n{traceback.format_exc()}")

    return wrapper


# Common Helper Functions


def get_long_type():
    """Return the cached long type object"""
    return long_type


def offset_of(typeobj: TypeOrStr, field: str) -> Union[int, None]:
    """Return the offset of a field in a structure"""
    if type(typeobj) is str:
        typeobj = gdb.lookup_type(typeobj)

    if typeobj.code is gdb.TYPE_CODE_PTR:
        typeobj = typeobj.target()

    for f in typeobj.fields():
        if f.name == field:
            if f.bitpos is None:
                break
            return f.bitpos // 8

    raise gdb.GdbError(f"Field {field} not found in type {typeobj}")


def container_of(ptr: Union[Value, int], typeobj: TypeOrStr, member: str) -> Value:
    """
    Return a pointer to the containing data structure.

    Args:
        ptr: Pointer to the member.
        t: Type of the container.
        member: Name of the member in the container.

    Returns:
        Value of the container.

    Example:
        struct foo {
            int a;
            int b;
        };
        struct foo *ptr = container_of(&ptr->b, "struct foo", "b");
    """

    if isinstance(typeobj, str):
        typeobj = lookup_type(typeobj).pointer()

    if typeobj.code is not gdb.TYPE_CODE_PTR:
        typeobj = typeobj.pointer()

    addr = Value(ptr).cast(long_type)
    return Value(addr - offset_of(typeobj, member)).cast(typeobj)


class ContainerOf(gdb.Function):
    """Return pointer to containing data structure.

    $container_of(PTR, "TYPE", "ELEMENT"): Given PTR, return a pointer to the
    data structure of the type TYPE in which PTR is the address of ELEMENT.
    Note that TYPE and ELEMENT have to be quoted as strings."""

    def __init__(self):
        super().__init__("container_of")

    def invoke(self, ptr, typename, elementname):
        return container_of(ptr, typename.string(), elementname.string())


ContainerOf()


class MacroCtx:
    """
    This is a singleton class which only initializes once to
    cache a context of macro definition which can be queried later
    TODO: we only deal with single ELF at the moment for simplicity
    If you load more object files while debugging, only the first one gets loaded
    will be used to retrieve macro information
    """

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "instance"):
            cls.instance = super(MacroCtx, cls).__new__(cls)
        return cls.instance

    def __init__(self, filename):
        self._macro_map = {}
        self._file = filename

        self._macro_map = fetch_macro_info(filename)

    @property
    def macro_map(self):
        return self._macro_map

    @property
    def objfile(self):
        return self._file


def parse_and_eval(expression: str, global_context: bool = False):
    """Equivalent to gdb.parse_and_eval, but returns a Value object"""
    gdb_value = gdb.parse_and_eval(expression)
    return Value(gdb_value)


def gdb_eval_or_none(expression):
    """Evaluate an expression and return None if it fails"""
    try:
        value = parse_and_eval(expression)
        if value.is_optimized_out:
            return None
        return value
    except gdb.error:
        return None


def suppress_cli_notifications(suppress=True):
    """Suppress(default behavior) or unsuppress GDB CLI notifications"""
    try:
        suppressed = "is on" in gdb.execute(
            "show suppress-cli-notifications", to_string=True
        )
        if suppress != suppressed:
            gdb.execute(f"set suppress-cli-notifications {'on' if suppress else 'off'}")

        return suppressed
    except gdb.error:
        return True


def check_inferior_valid():
    """Check if the current inferior is valid"""
    inferior = gdb.selected_inferior()
    return inferior and inferior.connection and inferior.connection.is_valid()


def get_symbol_value(name, locspec="nx_start", cacheable=True):
    """Return the value of a symbol value etc: Variable, Marco"""
    global g_symbol_cache

    # If there is a current stack frame, GDB uses the macros in scope at that frame’s source code line.
    # Otherwise, GDB uses the macros in scope at the current listing location.
    # Reference: https://sourceware.org/gdb/current/onlinedocs/gdb.html/Macros.html#Macros
    try:
        if not gdb.selected_frame():
            gdb.execute(f"list {locspec}", to_string=True)
            return gdb_eval_or_none(name)
    except gdb.error:
        pass

    # Try current frame
    value = gdb_eval_or_none(name)
    if value:
        return value

    # Check if the symbol is already cached
    if cacheable and (name, locspec) in g_symbol_cache:
        return g_symbol_cache[(name, locspec)]

    # There's current frame and no definition found. We need second inferior without a valid frame
    # in order to use the list command to set the scope.
    if len(gdb.inferiors()) == 1:
        gdb.execute(
            f'add-inferior -exec "{gdb.objfiles()[0].filename}" -no-connection',
            to_string=True,
        )
        g_symbol_cache = {}

    state = suppress_cli_notifications(True)

    original_thread = None
    original_frame = None
    if check_inferior_valid():
        original_thread = gdb.selected_thread()
        if original_thread and original_thread.is_valid():
            original_frame = gdb.selected_frame()
    # Switch to inferior 2 and set the scope firstly
    gdb.execute("inferior 2", to_string=True)
    gdb.execute(f"list {locspec}", to_string=True)
    value = gdb_eval_or_none(name)
    if not value:
        # Try to expand macro by reading elf
        global g_macro_ctx
        if not g_macro_ctx:
            if len(gdb.objfiles()) > 0:
                g_macro_ctx = MacroCtx(gdb.objfiles()[0].filename)
            else:
                raise gdb.GdbError("An executable file must be provided")

        expr = try_expand(name, g_macro_ctx.macro_map)
        value = gdb_eval_or_none(expr)

    if cacheable:
        g_symbol_cache[(name, locspec)] = value

    # Switch back to inferior 1
    gdb.execute("inferior 1", to_string=True)
    if original_thread and original_thread.is_valid():
        original_thread.switch()
        if original_frame and original_frame.is_valid():
            original_frame.select()
    suppress_cli_notifications(state)
    return value


def get_field(obj: gdb.Value, field: Union[str, gdb.Field], default=None) -> gdb.Value:
    """
    Get a field value from a gdb.Value, return default if field is not found.
    """
    try:
        return obj[field] if obj else default
    except gdb.error:
        return default


def has_field(obj: Union[TypeOrStr, gdb.Value], field):
    return get_type_field(obj, field) is not None


def get_bytes(val, size):
    """Convert a gdb value to a bytes object"""
    try:
        return val.bytes[:size]
    except AttributeError:  # Sometimes we don't have gdb.Value.bytes
        inf = gdb.inferiors()[0]
        mem = inf.read_memory(val.address, size)
        return mem.tobytes()


def import_check(module, name="", errmsg=""):
    try:
        module = __import__(module, fromlist=[name])
    except ImportError as e:
        gdb.write(f"import_error: {e}\n")
        gdb.write(errmsg if errmsg else f"Error to import {module}\n")
        return None

    return getattr(module, name) if name else module


def import_reload(module, name="", errmsg=""):
    """Clear and reload the module from sys.modules"""
    if module in sys.modules:
        module = importlib.reload(sys.modules[module])
    else:
        module = import_check(
            module,
            errmsg=errmsg,
        )
    return module


def hexdump(address, size):
    address = int(address)
    inf = gdb.inferiors()[0]
    mem = inf.read_memory(address, size)
    bytes = mem.tobytes()
    for i in range(0, len(bytes), 16):
        chunk = bytes[i : i + 16]
        gdb.write(f"{i + address:08x}  ")
        hex_values = " ".join(f"{byte:02x}" for byte in chunk)
        hex_display = f"{hex_values:<47}"
        gdb.write(hex_display)
        ascii_values = "".join(
            chr(byte) if 32 <= byte <= 126 else "." for byte in chunk
        )
        gdb.write(f"  {ascii_values} \n")


def is_decimal(s):
    return re.fullmatch(r"\d+", s) is not None


def is_hexadecimal(s):
    return re.fullmatch(r"0[xX][0-9a-fA-F]+|[0-9a-fA-F]+", s) is not None


def parse_arg(arg: str) -> Union[gdb.Value, int]:
    """Parse an argument to a gdb.Value or int, return None if failed"""

    if is_decimal(arg):
        return int(arg)

    if is_hexadecimal(arg):
        return int(arg, 16)

    try:
        return parse_and_eval(f"{arg}")
    except gdb.error:
        return None


def alias(name, command):
    try:
        gdb.execute(f"alias {name} = {command}")
    except gdb.error:
        pass


def nitems(array: Union[gdb.Field, gdb.Type, gdb.Symbol]) -> int:
    array_type = array.type
    element_type = array_type.target()
    element_size = element_type.sizeof
    array_size = array_type.sizeof // element_size
    return array_size


def sizeof(t: Union[str, gdb.Type]):
    if type(t) is str:
        t = gdb.lookup_type(t)

    return t.sizeof


def log2ceil(n):
    return int(math.ceil(math.log2(n)))


def log2floor(n):
    return int(math.floor(math.log2(n)))


# Machine Specific Helper Functions


BIG_ENDIAN = 0
LITTLE_ENDIAN = 1
target_endianness = None


def get_target_endianness():
    """Return the endianness of the target"""
    global target_endianness
    if not target_endianness:
        endian = gdb.execute("show endian", to_string=True)
        if "little endian" in endian:
            target_endianness = LITTLE_ENDIAN
        elif "big endian" in endian:
            target_endianness = BIG_ENDIAN
        else:
            raise gdb.GdbError("unknown endianness '{0}'".format(str(endian)))
    return target_endianness


def read_memoryview(inf, start, length):
    """Read memory from the target and return a memoryview object"""
    m = inf.read_memory(start, length)
    if type(m) is memoryview:
        return m
    return memoryview(m)


try:
    # For some prebuilt GDB, the python builtin module `struct` is not available
    import struct

    def read_u16(buffer, offset):
        """Read a 16-bit unsigned integer from a buffer"""
        if get_target_endianness() == LITTLE_ENDIAN:
            return struct.unpack_from("<H", buffer, offset)[0]
        else:
            return struct.unpack_from(">H", buffer, offset)[0]

    def read_u32(buffer, offset):
        """Read a 32-bit unsigned integer from a buffer"""
        if get_target_endianness() == LITTLE_ENDIAN:
            return struct.unpack_from("<I", buffer, offset)[0]
        else:
            return struct.unpack_from(">I", buffer, offset)[0]

    def read_u64(buffer, offset):
        """Read a 64-bit unsigned integer from a buffer"""
        if get_target_endianness() == LITTLE_ENDIAN:
            return struct.unpack_from("<Q", buffer, offset)[0]
        else:
            return struct.unpack_from(">Q", buffer, offset)[0]

except ModuleNotFoundError:

    def read_u16(buffer, offset):
        """Read a 16-bit unsigned integer from a buffer"""
        buffer_val = buffer[offset : offset + 2]
        value = [0, 0]

        if type(buffer_val[0]) is str:
            value[0] = ord(buffer_val[0])
            value[1] = ord(buffer_val[1])
        else:
            value[0] = buffer_val[0]
            value[1] = buffer_val[1]

        if get_target_endianness() == LITTLE_ENDIAN:
            return value[0] + (value[1] << 8)
        else:
            return value[1] + (value[0] << 8)

    def read_u32(buffer, offset):
        """Read a 32-bit unsigned integer from a buffer"""
        if get_target_endianness() == LITTLE_ENDIAN:
            return read_u16(buffer, offset) + (read_u16(buffer, offset + 2) << 16)
        else:
            return read_u16(buffer, offset + 2) + (read_u16(buffer, offset) << 16)

    def read_u64(buffer, offset):
        """Read a 64-bit unsigned integer from a buffer"""
        if get_target_endianness() == LITTLE_ENDIAN:
            return read_u32(buffer, offset) + (read_u32(buffer, offset + 4) << 32)
        else:
            return read_u32(buffer, offset + 4) + (read_u32(buffer, offset) << 32)


def read_uint(addr):
    buf = gdb.selected_inferior().read_memory(addr, 4)
    return int.from_bytes(buf, "little", signed=False)


def read_ulong(buffer, offset):
    """Read a long from a buffer"""
    if get_long_type().sizeof == 8:
        return read_u64(buffer, offset)
    else:
        return read_u32(buffer, offset)


def bswap(val, size):
    """Reverses the byte order in a gdb.Value or int value of size bytes"""
    return int.from_bytes(int(val).to_bytes(size, byteorder="little"), byteorder="big")


def swap16(val):
    return bswap(val, 2)


def swap32(val):
    return bswap(val, 4)


def swap64(val):
    return bswap(val, 8)


target_arch = None


def is_target_arch(arch, exact=False):
    """
    For non exact match, this function will
    return True if the target architecture contains
    keywords of an ARCH family. For example, x86 is
    contained in i386:x86_64.
    For exact match, this function will return True if
    the target architecture is exactly the same as ARCH.
    """
    if hasattr(gdb.Frame, "architecture"):
        archname = gdb.newest_frame().architecture().name()

        return arch in archname if not exact else arch == archname
    else:
        global target_arch
        if target_arch is None:
            target_arch = gdb.execute("show architecture", to_string=True)
            pattern = r'set to "(.*?)"\s*(\(currently (".*")\))?'
            match = re.search(pattern, target_arch)

            candidate = match.group(1)

            if candidate == "auto":
                target_arch = match.group(3)
            else:
                target_arch = candidate

        return arch in target_arch if not exact else arch == target_arch


# Kernel Specific Helper Functions

CONFIG_SMP_NCPUS = nitems(parse_and_eval("g_running_tasks"))


def is_target_smp():
    """Return Ture if the target use smp"""

    return CONFIG_SMP_NCPUS > 1


def get_ncpus():
    return CONFIG_SMP_NCPUS


# FIXME: support RISC-V/X86/ARM64 etc.
def in_interrupt_context(cpuid=0):
    frame = gdb.selected_frame()

    if is_target_arch("arm"):
        xpsr = int(frame.read_register("xpsr"))
        return xpsr & 0xF
    else:
        # TODO: figure out a more proper way to detect if
        # we are in an interrupt context
        g_current_regs = gdb_eval_or_none("g_current_regs")
        return not g_current_regs or not g_current_regs[cpuid]


# task
def get_register_byname(regname, tcb=None):
    frame = gdb.selected_frame()
    if not tcb:
        return int(frame.read_register(regname))

    if task_is_running(tcb):
        threads = gdb.selected_inferior().threads()
        if not threads:
            print("No threads found")
            return 0

        origin = gdb.selected_thread()

        # SMP online debugging is not currently supported by the debug tools.
        # However, in the case of QEMU, each simulated CPU runs in its own thread.
        # We can leverage this by inspecting QEMU threads in GDB to switch between CPUs,
        # and use GDB frames to access CPU registers.
        # This special case is handled below.
        if (
            check_inferior_valid()
            and "remote" in gdb.selected_inferior().connection.type.lower()
            and threads[0].details
            and "cpu" not in threads[0].details.lower()
        ):
            # We should use the TCB's pid to find the thread,
            # because gdbserver.py can parser the coredump
            thread = get_gdb_thread(tcb["pid"])
        else:
            # For SMP, we need to switch to the thread's CPU
            if is_target_smp() and len(threads) > 1:
                threads = sorted(threads, key=lambda t: t.num)
                thread = threads[tcb.cpu]
            else:
                # For single CPU, just use the first thread
                thread = threads[0]
        if thread and thread.is_valid():
            thread.switch()
            reg = int(gdb.selected_frame().read_register(regname))
            if origin and origin.is_valid():
                origin.switch()
                if frame and frame.is_valid():
                    frame.select()
            else:
                print("Warning: the original thread is not valid")
            return reg
        else:
            print(f"Thread is not valid, tcb: {tcb}")
            return 0

    # Ok, let's take it from the context in the given tcb
    tcbinfo = parse_and_eval("g_tcbinfo")
    for reg in ArrayIterator(tcbinfo.u.reginfo, tcbinfo.regs_num):
        if reg.name.string().lower() == regname.lower():
            xcpregs = tcb["xcp"]["regs"]
            if xcpregs.type.code != gdb.TYPE_CODE_PTR:
                # For sim, the xcpregs is an array, not pointer, thus need casting
                xcpregs = xcpregs.cast(lookup_type("char").pointer())
            value = gdb.Value(int(xcpregs) + reg.toffset)
            value = value.cast(lookup_type("uintptr_t").pointer())
            try:
                return int(value.dereference())
            except gdb.MemoryError:
                print(
                    f"Failed to read memory at {value}, tcb: {tcb['pid']}\n"
                    "Possible reason: The task is going to run soon, so tcb.xcp.regs is null"
                )
                return 0


def get_sp(tcb=None):
    # NuttX doesn't support unified register name for stack pointer
    arch = gdb.selected_inferior().architecture().name()
    regname = {
        "i386": "esp",
        "i386:x86": "esp",
        "i386:x86-64": "rsp",
    }.get(arch, "sp")

    return get_register_byname(regname, tcb)


def get_pc(tcb=None):
    # NuttX doesn't support unified register name for PC
    arch = gdb.selected_inferior().architecture().name()
    regname = {
        "i386": "eip",
        "i386:x86": "eip",
        "i386:x86-64": "rip",
    }.get(arch, "pc")
    return get_register_byname(regname, tcb)


def get_tcbs() -> List[Tcb]:
    # In case we have created/deleted tasks at runtime, the tcbs will change
    # so keep it as fresh as possible
    pidhash = parse_and_eval("g_pidhash")
    npidhash = parse_and_eval("g_npidhash")

    return [pidhash[i] for i in range(0, npidhash) if pidhash[i]]


def get_tcb(pid) -> Tcb:
    """get tcb from pid"""
    pidhash = parse_and_eval("g_pidhash")
    npidhash = parse_and_eval("g_npidhash")
    tcb = pidhash[pid & (npidhash - 1)]
    if not tcb or pid != tcb["pid"]:
        return None

    return tcb


def get_running_tcbs() -> List[Tcb]:
    """get running tcbs"""
    running_tasks = parse_and_eval("g_running_tasks")
    return [tcb for tcb in ArrayIterator(running_tasks) if tcb]


def get_tcb_type(tcb):
    """get tcb type"""
    if not tcb:
        return None
    try:
        mask = get_symbol_value("TCB_FLAG_TTYPE_MASK")
        if tcb["flags"] & mask == get_symbol_value("TCB_FLAG_TTYPE_PTHREAD"):
            return "PTHREAD"
        elif tcb["flags"] & mask == get_symbol_value("TCB_FLAG_TTYPE_KERNEL"):
            return "KTHREAD"
        elif tcb["flags"] & mask == get_symbol_value("TCB_FLAG_TTYPE_TASK"):
            return "TASK"
        return None
    except gdb.error as e:
        raise gdb.error(f"Failed to determine TCB type: {str(e)}") from e


def get_tid(tcb):
    """get tid from tcb"""
    if not tcb:
        return None
    try:
        return tcb["group"]["tg_pid"]
    except gdb.error:
        return None


def get_task_name(tcb_or_pid):
    if isinstance(tcb_or_pid, int):
        tcb = get_tcb(tcb_or_pid)
    else:
        tcb = tcb_or_pid

    if tcb is None:
        return ""

    try:
        name = tcb["name"].cast(gdb.lookup_type("char").pointer())
        return name.string()
    except gdb.error:
        return ""


def get_task_entry(tcb) -> Symbol:
    if not tcb:
        return None

    try:
        entry = tcb["entry"]["main"]
        return Symbol(entry)
    except gdb.error:
        return None


def task_is_running(tcb):
    return tcb["task_state"] == get_symbol_value("TSTATE_TASK_RUNNING")


def task_is_idle(tcb):
    return tcb["pid"] < CONFIG_SMP_NCPUS


# sem
def sem_is_mutex(sem):
    SEM_TYPE_MUTEX = 4
    return sem["flags"] & SEM_TYPE_MUTEX


def mutex_get_holder(mutex):
    NXSEM_MBLOCKING_BIT = 1 << 31
    return mutex["val"]["mholder"] & ~NXSEM_MBLOCKING_BIT


def switch_inferior(inferior):
    state = suppress_cli_notifications(True)

    if len(gdb.inferiors()) == 1:
        gdb.execute(
            f"add-inferior -exec {gdb.objfiles()[0].filename} -no-connection",
            to_string=True,
        )

    gdb.execute(f"inferior {inferior}", to_string=True)
    return state


def check_version():
    """Check the elf and memory version"""
    original_thread = None
    original_frame = None
    state = suppress_cli_notifications()
    if check_inferior_valid():
        original_thread = gdb.selected_thread()
        if original_thread and original_thread.is_valid():
            original_frame = gdb.selected_frame()
    switch_inferior(1)
    try:
        mem_version = gdb.execute("p g_version", to_string=True).split("=")[1]
    except gdb.error:
        gdb.write("No symbol g_version found in memory, skipping version check\n")
        suppress_cli_notifications(state)
        return

    switch_inferior(2)
    elf_version = gdb.execute("p g_version", to_string=True).split("=")[1]
    if mem_version != elf_version:
        gdb.write(f"\x1b[31;1mMemory version:{mem_version}")
        gdb.write(f"ELF version:   {elf_version}")
        gdb.write("Warning version not matched, please check!\x1b[m\n")
    else:
        gdb.write(f"Build version: {mem_version}\n")

    switch_inferior(1)  # Switch back
    if original_thread and original_thread.is_valid():
        original_thread.switch()
        if original_frame and original_frame.is_valid():
            original_frame.select()
    # Verify the ELF file version against the GDB tool version
    check_compatibility()
    suppress_cli_notifications(state)


def check_compatibility():
    """Check the elf and the GDB tool version"""
    from . import uname

    kernel_version = uname.kernel_version
    version_parts = kernel_version.split(" ")
    if len(version_parts) < 1:
        gdb.write(f"Invalid kernel_version format: {kernel_version}")
        return
    elf_version = version_parts[0]
    tool_verson = uname.tool_version

    if not elf_version.startswith(tool_verson):
        gdb.write(
            f"\x1b[31;1mWarning: ELF version {elf_version} "
            f"does not match GDB tool version {tool_verson}\x1b[m\n"
        )


def get_task_tls(tid, key):
    """get task tls from tid and key"""
    tcb = get_tcb(tid)
    if not tcb:
        return None

    try:
        stack_alloc_ptr = tcb["stack_alloc_ptr"].cast(
            lookup_type("struct tls_info_s").pointer()
        )
        tls_value = stack_alloc_ptr["tl_task"]["ta_telem"][int(key)]
        return tls_value.cast(lookup_type("uintptr_t").pointer())
    except gdb.error:
        return None


def get_thread_tls(pid, key):
    """get thread tls from pid and key"""
    tcb = get_tcb(pid)
    if not tcb:
        return None

    try:
        stack_alloc_ptr = tcb["stack_alloc_ptr"].cast(
            lookup_type("struct tls_info_s").pointer()
        )
        tls_value = stack_alloc_ptr["tl_elem"][int(key)]
        return tls_value.cast(lookup_type("uintptr_t").pointer())
    except gdb.error:
        return None


def get_task_argvstr(tcb: Tcb) -> List[str]:
    args = []
    try:
        TCB_FLAG_TTYPE_MASK = get_symbol_value("TCB_FLAG_TTYPE_MASK")
        TCB_FLAG_TTYPE_PTHREAD = get_symbol_value("TCB_FLAG_TTYPE_PTHREAD")

        if tcb.flags & TCB_FLAG_TTYPE_MASK == TCB_FLAG_TTYPE_PTHREAD:
            if tcb.type.code != gdb.TYPE_CODE_PTR:
                tcb = tcb.address
            ptcb = tcb.cast(lookup_type("struct pthread_entry_s").pointer())
            return ["", f"{tcb['entry']['main']}", f'{ptcb["arg"]}']

        tls_info_s = lookup_type("struct tls_info_s").pointer()
        tls = tcb.stack_alloc_ptr.cast(tls_info_s)
        argv = int(tcb.stack_alloc_ptr) + int(tls.tl_size)
        argv = gdb.Value(argv).cast(lookup_type("char").pointer().pointer())
        while argv.dereference():
            args.append(argv.dereference().string())
            argv += 1
    except gdb.error:
        pass

    return args


def gather_modules(dir=None) -> List[str]:
    dir = Path(dir).resolve() if dir else Path(__file__).parent
    modules = []

    for f in dir.rglob("*.py"):
        if f.name == "__init__.py":
            continue

        relative_path = f.relative_to(dir).with_suffix("")
        module_name = ".".join(relative_path.parts)
        modules.append(module_name)

    return modules


def gather_gdbcommands(modules=None, path=None) -> List[gdb.Command]:
    modules = modules or gather_modules(path)
    commands = []
    for m in modules:
        try:
            module = importlib.import_module(f"{__package__}.{m}")
            for c in module.__dict__.values():
                if isinstance(c, type) and issubclass(c, gdb.Command):
                    commands.append(c)
        except Exception:
            gdb.write(f"Ignore module {m}\n")
    return commands


def get_elf_md5():
    """Return the md5 checksum of the current ELF file"""
    file = gdb.objfiles()[0].filename
    with open(file, "rb") as f:
        hash = hashlib.md5(f.read()).hexdigest()
    return hash


def jsonify(obj, indent=None):
    if not obj:
        return "{}"

    def dumper(obj):
        try:
            return str(obj) if isinstance(obj, gdb.Value) else obj.toJSON()
        except Exception:
            try:
                return obj.__dict__
            except Exception as e:
                return f"<jsonify failed: {type(obj), {e}}>"

    return json.dumps(obj, default=dumper, indent=indent)


def enum(t: Union[str, gdb.Type], name=None):
    """Create python Enum class from C enum values
    Usage:

    in C:
    enum color_e {
        RED = 1,
        GREEN = 2,
    };

    in python:
    COLOR = utils.enum("enum color_e", "COLOR")
    print(COLOR.GREEN.value) # --> 2
    RED = COLOR(1)
    """
    if type(t) is str:
        t = lookup_type(t) or lookup_type("enum " + t)

    if t and t.code == gdb.TYPE_CODE_TYPEDEF:
        t = t.strip_typedefs()

    if not t or t.code != gdb.TYPE_CODE_ENUM:
        raise gdb.error(f"{t} is not an enum type")

    def commonprefix(m):
        "Given a list of pathnames, returns the longest common leading component"
        if not m:
            return ""
        s1 = min(m)
        s2 = max(m)
        for i, c in enumerate(s1):
            if c != s2[i]:
                return s1[:i]
        return s1

    # Remove the common prefix from names. This is a convention in python.
    # E.g. COLOR.RED, COLOR.GREEN instead of COLOR.COLOR_RED, COLOR.COLOR_GREEN

    prefix = commonprefix([f.name for f in t.fields()])

    names = {f.name[len(prefix) :]: f.enumval for f in t.fields()}

    if not name:
        if prefix:
            name = prefix[:-1] if prefix[-1] == "_" else prefix
        else:
            name = "Enum"

    return Enum(name, names)


class ArrayIterator:
    """An iterator for gdb array or pointer."""

    def __init__(self, array: gdb.Value, maxlen=None, reverse=False):
        type_code = array.type.code
        if type_code not in (gdb.TYPE_CODE_ARRAY, gdb.TYPE_CODE_PTR):
            raise gdb.error(f"Not an array: {array}, type: {array.type}")

        if type_code == gdb.TYPE_CODE_ARRAY:
            if (n := nitems(array)) > 0:
                maxlen = min(n, maxlen) if maxlen is not None else n

        if maxlen is None:
            raise gdb.error("Need to provide array length.")

        self.array = array
        self.maxlen = maxlen
        self.reverse = reverse
        self.index = maxlen - 1 if reverse else 0

    def __iter__(self):
        return self

    def __next__(self) -> gdb.Value:
        if (not self.reverse and self.index >= self.maxlen) or (
            self.reverse and self.index < 0
        ):
            raise StopIteration

        value = self.array[self.index]
        self.index = self.index - 1 if self.reverse else self.index + 1
        return value


class Hexdump(gdb.Command):
    """hexdump address/symbol <size>"""

    def __init__(self):
        super().__init__("hexdump", gdb.COMMAND_USER)

    @dont_repeat_decorator
    def invoke(self, args, from_tty):
        argv = args.split(" ")
        address = 0
        size = 0
        if argv[0] == "":
            gdb.write("Usage: hexdump address/symbol <size>\n")
            return

        if is_decimal(argv[0]) or is_hexadecimal(argv[0]):
            address = int(argv[0], 0)
            size = int(argv[1], 0)
        else:
            try:
                var = gdb.parse_and_eval(f"{argv[0]}")
                address = int(var.cast(long_type))
                size = int(argv[1]) if argv[1] else int(var.type.sizeof)
                gdb.write(f"{argv[0]} {hex(address)} {int(size)}\n")
            except Exception as e:
                gdb.write(f"Invalid {argv[0]}: {e}\n")

        hexdump(address, size)


PID0_REPLACE = 0x7FFFFFFF
MAX_FRAMES = 99


def get_gdb_thread_pid(thread: gdb.InferiorThread) -> int:
    return 0 if thread.ptid[1] == PID0_REPLACE else thread.ptid[1]


def get_gdb_thread(pid: int) -> Optional[gdb.InferiorThread]:
    for thread in gdb.selected_inferior().threads():
        if get_gdb_thread_pid(thread) == pid:
            return thread
    return None


def get_thread_frames(arg: Union[gdb.InferiorThread, int]) -> Union[List[gdb.Frame]]:
    thread = arg if not isinstance(arg, int) else get_gdb_thread(arg)
    if not thread:
        return []

    thread.switch()

    frames = []
    frame = gdb.newest_frame()

    while len(frames) < MAX_FRAMES and frame and frame.is_valid():
        frames.append(frame)
        frame = frame.older()

    return frames


def get_backtrace(arg: Union[gdb.InferiorThread, int]) -> List[int]:
    backtrace = []

    for frame in get_thread_frames(arg):
        backtrace.append(frame.pc())

    return backtrace


def get_frame_func_name(frame: gdb.Frame) -> str:
    function = frame.function()
    if function and function.is_valid():
        return function.name

    return ""


def get_frame_variables(frame: gdb.Frame) -> dict:
    varibles = {}
    try:
        block = frame.block()
        while block and block.is_valid():
            for symbol in block:
                if symbol.is_valid() and symbol.is_variable:
                    varibles[symbol.name] = symbol.value(frame)
            block = block.superblock
    except RuntimeError:
        pass

    return varibles



############################################################################
# tools/pynuttx/nxgdb/__init__.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import importlib
import traceback
from os import path

import gdb

if __name__ == "__main__":
    import sys

    gdb.write("Use nuttx/tools/pynuttx/gdbinit.py instead")
    sys.exit(1)

here = path.dirname(path.abspath(__file__))


def register_commands(event):
    if getattr(register_commands, "registered", False):
        return

    register_commands.registered = True

    gdb.write(f"Registering NuttX GDB commands from {here}\n")
    gdb.execute("set pagination off")
    gdb.write("set pagination off\n")
    gdb.execute("set python print-stack full")
    gdb.write("set python print-stack full\n")
    gdb.execute('handle SIGUSR1 "nostop" "pass" "noprint"')
    gdb.write('"handle SIGUSR1 "nostop" "pass" "noprint"\n')

    def init_gdb_commands(m: str):
        try:
            module = importlib.import_module(f"{__package__}.{m}")
        except Exception as e:
            gdb.write(f"\x1b[31;1mIgnore module: {m}, error: {e}\n\x1b[m")
            traceback.print_exc()
            return

        for c in module.__dict__.values():
            if isinstance(c, type) and issubclass(c, gdb.Command):
                try:
                    c()
                except Exception as e:
                    gdb.write(f"\x1b[31;1mIgnore command: {c}, e: {e}\n\x1b[m")

    # Import GDBVenv firstly, so we can always create venv to install requirements for pynuttx.
    venv_cmd = importlib.import_module(f"{__package__}.venv")
    venv_cmd.GDBVenv()

    # import utils module
    utils = importlib.import_module(f"{__package__}.utils")
    modules = utils.gather_modules(here)

    # Register prefix commands firstly
    init_gdb_commands("prefix")
    modules.remove("prefix")

    # Register all other modules
    for m in modules:
        init_gdb_commands(m)

    utils.check_version()


if len(gdb.objfiles()) == 0:
    gdb.write("No objectfile, defer to register NuttX commands.\n")
    gdb.events.new_objfile.connect(register_commands)
else:
    register_commands(None)

print(
    "Hint: use `gdbvenv` to create virtual environment for nxgdb, "
    "or `gdbvenv --install [path/to/requirements.txt]` to reinstall the packages"
)



############################################################################
# tools/pynuttx/nxgdb/notesnap.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import gdb

from . import utils

CONFIG_SMP = utils.has_field("struct notesnap_chunk_s", "cpu")
CONFIG_DRIVERS_NOTESNAP_NBUFFERS = utils.get_field_nitems("struct notesnap_s", "buffer")


class NoteSnapChunk:
    def __init__(self, notetype, pid, args, cpu=0):
        self.notetype = notetype
        self.pid = pid
        self.args = args
        self.cpu = cpu

    @property
    def taskname(self):
        return utils.get_task_name(utils.get_tcb(self.pid))

    def __str__(self):
        args = [
            self.taskname,
        ]
        if CONFIG_SMP:
            args.append(self.cpu)
        args.extend([self.pid, self.notetype, self.args])
        return (
            "[%s] " + ("[CPU%d] " if CONFIG_SMP else "") + "[%d] %-16s %#x"
        ) % tuple(args)


class NoteSnap(gdb.Command):
    """Dump notesnap"""

    def __init__(self):
        if not CONFIG_DRIVERS_NOTESNAP_NBUFFERS:
            return
        super().__init__("notesnap", gdb.COMMAND_USER)

    def parse_notesnap(self):
        """Parse note snapshot data"""

        notesnap = utils.parse_and_eval("g_notesnap")
        notesnap_type = utils.parse_and_eval("g_notesnap_type")
        output = []

        for note in utils.ArrayIterator(notesnap.buffer):
            if not note:
                continue

            chunk = {
                "notetype": notesnap_type[note["type"]].string(),
                "pid": note.pid,
                "args": note.args,
                "cpu": note.cpu if CONFIG_SMP else 0,
            }

            output.append(NoteSnapChunk(**chunk))
        return output

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        try:
            chunks = self.parse_notesnap()

            prettytable = utils.import_check(
                "prettytable",
                errmsg="Execute `pip install prettytable` for better printing result.\n",
            )

            if prettytable:
                table = prettytable.PrettyTable()
                headers = ["Task Name", "PID", "Note Type", "Args"]
                if CONFIG_SMP:
                    headers.insert(1, "CPU")
                table.field_names = headers
                table.align = "l"
                for chunk in chunks:
                    row = [chunk.taskname]
                    if CONFIG_SMP:
                        row.append(str(chunk.cpu))
                    row.extend([str(chunk.pid), chunk.notetype, hex(chunk.args)])
                    table.add_row(row)
                print(table)
            else:
                print(*chunks, sep="\n")

        except Exception as e:
            print(f"Error happened during notesnap: {e}")

    def diagnose(self, *args, **kwargs):
        return {
            "title": "Notesnap Report",
            "summary": "notesnap dump",
            "command": "notesnap",
            "result": "info",
            "category": utils.DiagnoseCategory.system,
            "message": self.parse_notesnap(),
        }



############################################################################
# tools/pynuttx/nxtrace/rtt.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
import os
import subprocess
import threading
import time

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARN)


class SeggerRTT:
    def __init__(
        self,
        device,
        interface,
        address=None,
        speed=None,
        channel=None,
    ):
        self.receive_total = 0
        self.receive_rate = 0

        # Check if the JLinkRTTLogger command exists
        if not any(
            os.access(os.path.join(path, "JLinkRTTLogger"), os.X_OK)
            for path in os.environ["PATH"].split(os.pathsep)
        ):
            raise RuntimeError(
                "JLinkRTTLogger command not found, please ensure SEGGER JLink software is installed"
            )

        if channel is None:
            channel = 0
        if speed is None:
            speed = 16000

        # Create a named pipe
        self.pipe = f"/tmp/jlink_rtt_pipe_{channel}"
        if not os.path.exists(self.pipe):
            os.mkfifo(self.pipe)

        # Start the JLinkRTTLogger process
        cmd = [
            "JLinkRTTLogger",
            "-If",
            interface,
            "-device",
            device,
            "-RTTChannel",
            f"{channel}",
            "-Speed",
            f"{speed}",
        ]

        if address is not None:
            cmd.append("-RTTAddress")
            cmd.append(f"0x{address:x}")

        cmd.append(self.pipe)
        self.running = False
        self.buffer = bytearray()
        self.condition = threading.Condition()
        print(f"Starting JLinkRTTLogger with command: {cmd}")
        self.jlink_process = subprocess.Popen(cmd)
        self.thread = threading.Thread(target=self.process_data, daemon=True)

    def __del__(self):
        if self.running:
            self.stop()

    def process_data(self):
        last_receive_total = 0
        last_time = time.time()

        # Independent thread receives data to avoid data loss
        with open(self.pipe, "rb") as f:
            while self.running:
                data = f.read1()

                if data is None:
                    continue

                if time.time() - last_time > 0.5:
                    now_time = time.time()
                    self.receive_rate = (self.receive_total - last_receive_total) / (
                        now_time - last_time
                    )
                    last_time = now_time
                    last_receive_total = self.receive_total
                    logger.debug(
                        f"Receive rate: {self.receive_rate/1024:.2f} KB/s, total: {self.receive_total/1024/1024:.2f} MB"
                    )
                    self.receive_total += len(data)

                # Notify the waiting thread that data is available
                with self.condition:
                    self.buffer.extend(data)
                    self.condition.notify()

    def read(self, size=1024):
        try:
            with self.condition:
                while not self.buffer:
                    # Wait for data to be available
                    self.condition.wait()

            data = bytes(self.buffer[:size])
            self.buffer = self.buffer[size:]
            return data
        except Exception as e:
            logger.error(f"Error reading data: {e}")
            exit(1)

    def start(self):
        if self.jlink_process.poll() is not None:
            logger.error("Failed to start JLinkRTTLogger process")
            raise RuntimeError("Failed to start JLinkRTTLogger process")
        self.running = True
        self.thread.start()

    def stop(self):
        self.running = False
        self.thread.join()
        os.unlink(self.pipe)



############################################################################
# tools/pynuttx/nxgdb/elf.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
import os

import gdb

from . import autocompeletion

try:
    import lief
except ImportError:
    print('Package missing, please do "pip install lief"')

from .utils import dont_repeat_decorator, get_tcbs, has_field

CONFIG_ARCH_USE_SEPARATED_SECTION = has_field("struct module_s", "sectalloc")


@autocompeletion.complete
class ElfImport(gdb.Command):

    def get_argparser(self):
        parser = argparse.ArgumentParser(description="import elf symbols to gdb")
        parser.add_argument(
            "elfpath", type=str, metavar="file", help="elf file path, etc: apps/bin"
        )
        return parser

    def __init__(self):
        if has_field("struct task_group_s", "tg_bininfo"):
            super().__init__("elfimport", gdb.COMMAND_USER)
            self.parser = self.get_argparser()

    @dont_repeat_decorator
    def invoke(self, args, from_tty):
        try:
            args = self.parser.parse_args(args.split())
        except SystemExit:
            return None

        tcbs = get_tcbs()
        if tcbs is None:
            print("No TCBs found")
            return
        modules = []
        for tcb in tcbs:
            if (
                tcb.group != 0
                and tcb.group.tg_bininfo != 0
                and tcb.group.tg_bininfo.mod.modname.string() != ""
            ):
                modules.append(tcb.group.tg_bininfo.mod)

        if not modules:
            print("No modules in current environment")
            return

        print(f"Found {len(modules)} modules in current environment")

        for mod in modules:
            sections = {}
            cmd = ""
            if CONFIG_ARCH_USE_SEPARATED_SECTION:
                for i in range(mod.nsect):
                    section = mod.sectalloc[i]
                    if section:
                        sections[i] = hex(section)
                filename = os.path.join(args.elfpath, mod.modname.string())
                cmd = f"add-symbol-file {filename} "
                elf = lief.parse(filename)
                for i, section in sections.items():
                    cmd += f"-s {elf.sections[i].name} {sections[i]} "
            else:
                filename = os.path.join(args.elfpath, mod.modname.string())
                cmd = f"add-symbol-file {filename} -s .text {hex(mod.textalloc)} -s .data {hex(mod.dataalloc)}"

            print(cmd)
            gdb.execute(cmd)



#!/usr/bin/python3
############################################################################
# tools/pynuttx/gdbrpc/__main__.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse

if __name__ == "__main__":
    import gdbrpc

    parser = argparse.ArgumentParser(description="GDB Socket Client")
    parser.add_argument("--host", default="localhost", help="Server host")
    parser.add_argument("--port", type=int, default=20819, help="Server port")

    args = parser.parse_args()

    client = gdbrpc.ClientCLI(args.host, args.port)
    client.start()



############################################################################
# tools/pynuttx/nxgdb/pginfo.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import gdb

from .utils import DiagnoseCategory, dont_repeat_decorator, parse_and_eval, read_u64


class PageTable:
    """Base class for page table handling."""

    def __init__(self):
        pass

    def walk_table(self, base, level, prefix=""):
        """Placeholder for architecture-specific table walking logic."""
        raise NotImplementedError("Subclasses must implement this method.")

    def dump(self):
        """Placeholder for architecture-specific dumping logic."""
        raise NotImplementedError("Subclasses must implement this method.")


class PageTableX86(PageTable):
    """x86_64-specific page table walker."""

    PAGE_SIZE = 4096
    NUM_ENTRIES = 512
    ENTRY_SIZE = 8
    PAGE_PRESENT = 1 << 0
    PAGE_WRITABLE = 1 << 1
    PAGE_PSE = 1 << 7

    PAGE_LEVELS = {0: "PML4", 1: "PDPT", 2: "PD", 3: "PT"}
    PAGE_SIZES = {1: "1G", 2: "2M", 3: "4K"}

    def __init__(self):
        super().__init__()

    def walk_table(self, base, level, prefix=""):
        """x86_64-specific implementation to traverse the page table and yield entries."""
        if level > 3:
            return

        table_name = self.PAGE_LEVELS.get(level, f"L{level}")
        for i in range(self.NUM_ENTRIES):
            entry_addr = base + i * self.ENTRY_SIZE + 0x100000000
            data = gdb.selected_inferior().read_memory(entry_addr, self.ENTRY_SIZE)
            entry = read_u64(data, 0)

            if not (entry & self.PAGE_PRESENT):
                continue

            addr = entry & ~0xFFF
            flags = "RW" if entry & self.PAGE_WRITABLE else "R"
            if level == 3 or (level in (2, 1) and entry & self.PAGE_PSE):
                size = self.PAGE_SIZES.get(level, "")
            else:
                size = ""

            if size:
                line = f"{prefix}[{table_name} {i}] -> {addr:#x} ({flags} {size})"
            else:
                line = f"{prefix}[{table_name} {i}] -> {addr:#x} ({flags})"
            yield line

            if level < 3 and not (level in (2, 1) and entry & self.PAGE_PSE):
                yield from self.walk_table(addr, level + 1, prefix + "    ")

            if level == 0:
                break

    def dump(self):
        pml4_addr = int(parse_and_eval("&pml4")) & ~0xFFF
        print(f"pml4_addr: {pml4_addr:#x}")

        for line in self.walk_table(pml4_addr, 0):
            print(line)


class DumpPageTableCommand(gdb.Command):
    """GDB command to dump page tables based on architecture."""

    def __init__(self):
        super().__init__("dump_pagetable", gdb.COMMAND_USER)

    @dont_repeat_decorator
    def invoke(self, arg, from_tty):

        arch = gdb.selected_inferior().architecture().name()
        arch_classes = {
            "i386:x86-64": PageTableX86,  # Add other architectures as needed
        }

        clz = arch_classes.get(arch)
        if not clz:
            print(f"Unsupported architecture: {arch}")
            return

        pagetable = clz()
        pagetable.dump()

    def diagnose(self, *args, **kwargs):
        return {
            "title": "Page Table Information",
            "summary": "Page table information",
            "command": "dump_pagetable",
            "result": "info",
            "category": DiagnoseCategory.memory,
            "message": gdb.execute("dump_pagetable", to_string=True),
        }



############################################################################
# tools/pynuttx/nxgdbmcp/src/gmcp/tools/value.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from typing import Optional, Union

from mcp.server.fastmcp import Context

from ..context import get_session
from ..utils import _exec_command, error_handler


@error_handler
async def _examine(
    ctx: Context,
    session_id: str,
    expression: Union[str, int],
    format: str = "x",
    count: int = 1,
) -> str:
    session = get_session(ctx, session_id)

    # Convert int to string for GDB
    expr_str = str(expression)

    # Map format codes to GDB format specifiers
    format_map = {
        "x": "x",  # hex
        "d": "d",  # decimal
        "u": "u",  # unsigned decimal
        "o": "o",  # octal
        "t": "t",  # binary
        "i": "i",  # instruction
        "c": "c",  # char
        "f": "f",  # float
        "s": "s",  # string
    }

    gdb_format = format_map.get(format, "x")
    command = f"x/{count}{gdb_format} {expr_str}"
    output = await session.execute_command(command)
    return f"Examine {expr_str} (format: {format}, count: {count}):\n\n{output}"


@error_handler
async def _watchpoint(
    ctx: Context,
    session_id: str,
    expression: Union[str, int],
    watch_type: str = "write",
) -> str:
    session = get_session(ctx, session_id)

    # Convert int to string for GDB
    expr_str = str(expression)

    # Map watch types to GDB options
    watch_options = {"read": "r", "write": "w", "read_write": "aw"}
    option = watch_options.get(watch_type, "w")

    if option == "r":
        output = await session.execute_command(f"rwatch {expr_str}")
    elif option == "aw":
        output = await session.execute_command(f"awatch {expr_str}")
    else:
        output = await session.execute_command(f"watch {expr_str}")
    return f"Watchpoint set on {expr_str} (type: {watch_type})\n\nOutput:\n{output}"


def register_value_tools(gdb_mcp):
    @gdb_mcp.tool()
    async def gdb_print(
        ctx: Context, session_id: str, expression: Union[str, int]
    ) -> str:
        """Print value of expression.

        Args:
            session_id: The GDB session identifier
            expression: A string containing the GDB expression to evaluate
                (e.g., "variable_name", "*ptr", "array[0]"),
                or an integer representing a decimal memory address
        """
        expr_str = str(expression)
        return await _exec_command(ctx, session_id, f"print {expr_str}")

    @gdb_mcp.tool()
    async def gdb_examine(
        ctx: Context,
        session_id: str,
        expression: Union[str, int],
        format: str = "x",
        count: int = 1,
    ) -> str:
        """Examine memory at the specified address or expression.

        Args:
            session_id: The GDB session identifier
            expression: A string containing the memory address or expression
                (e.g., "0x12345678", "&variable", "ptr"),
                or an integer representing a decimal memory address
            format: Display format:
                "x" (hex)
                "d" (decimal)
                "u" (unsigned)
                "o" (octal)
                "t" (binary)
                "i" (instruction)
                "c" (char)
                "f" (float)
                "s" (string)
            count: Number of items to display
        """
        return await _examine(ctx, session_id, expression, format, count)

    @gdb_mcp.tool()
    async def gdb_info_registers(
        ctx: Context, session_id: str, register: Optional[str] = None
    ) -> str:
        """Display CPU register values.

        Args:
            session_id: The GDB session identifier
            register: Optional register name as a string
                (e.g., "rax", "rsp", "pc"). If None, displays all registers.
        """
        command = "info registers"
        command += f" {register}" if register is not None else ""
        return await _exec_command(ctx, session_id, command)

    @gdb_mcp.tool()
    async def gdb_watchpoint(
        ctx: Context,
        session_id: str,
        expression: Union[str, int],
        watch_type: str = "write",
    ) -> str:
        """Set a watchpoint on a variable or memory address.

        Args:
            session_id: The GDB session identifier
            expression: A string containing the variable name or memory address to watch
                (e.g., "my_variable", "*0x12345678"),
                or an integer representing a decimal memory address
            watch_type: Type of watchpoint - "write" (default), "read", or "read_write"
        """
        return await _watchpoint(ctx, session_id, expression, watch_type)

    @gdb_mcp.tool()
    async def gdb_expression(
        ctx: Context, session_id: str, expression: Union[str, int]
    ) -> str:
        """Evaluate an expression in the current frame.

        Args:
            session_id: The GDB session identifier
            expression: A string containing the expression to evaluate
                (e.g., "x + y", "func(arg)", "sizeof(struct)"),
                or an integer representing a decimal memory address
        """
        expr_str = str(expression)
        return await _exec_command(ctx, session_id, f"print {expr_str}")



############################################################################
# tools/pynuttx/nxgdb/uorb.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from __future__ import annotations

import argparse
import re
from typing import List

import gdb
import nxgdb.fs as fs

from . import autocompeletion, utils
from .circbuf import CircBuf
from .protocols import uorb as p


class Sensor(utils.Value, p.SensorUpper):
    """struct sensor_upperhalf_s and enhancement"""

    inode_s = utils.lookup_type("struct inode")
    sensor_upperhalf_s = utils.lookup_type("struct sensor_upperhalf_s")

    def __init__(self, inode: gdb.Value | utils.Value, path=None):
        # inode must be type of struct inode *
        if inode.type.code != gdb.TYPE_CODE_PTR:
            raise ValueError(f"Expect pointer type, got {inode.type}")

        super().__init__(inode["i_private"].cast(self.sensor_upperhalf_s.pointer()))
        self.inode = inode
        self._path = path

    def __repr__(self) -> str:
        state = self.state
        return f"{hex(self)} {self.topicname} {state.nsubscribers} subscribers, {state.nadvertisers} advertisers"

    def details(self) -> str:
        state = self.state
        return f"nbuffer: {state.nbuffer}, latency: {state.min_latency}, interval: {state.min_interval}"

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def path(self):
        return self._path or fs.inode_getpath(self.inode)

    @property
    def nsubscribers(self) -> int:
        return int(self.state["nsubscribers"])

    @property
    def nadvertisers(self) -> int:
        return int(self.state["nadvertisers"])

    @property
    def topicname(self):
        name = self.path.split("/")[-1]
        name = re.sub(r"(\d$)", "", name)
        name = re.sub(r"(_uncal$)", "", name)
        return name

    @property
    def metadata(self) -> p.OrbMetadata:
        return utils.gdb_eval_or_none(f"g_orb_{self.topicname}")

    @property
    def datatype(self) -> gdb.Type:
        """Return the datatype of the topic like struct sensor_accel"""
        return utils.lookup_type(f"struct {self.topicname}")

    @property
    def circbuf(self) -> CircBuf:
        if not self.datatype:
            return None
        return CircBuf(self.buffer, datatype=self.datatype.pointer())


def get_topic_inodes(topic: str = None) -> List[fs.Inode]:
    nodes = (
        (node, path)
        for node, path in fs.foreach_inode()
        if path.startswith("/dev/uorb/") and (not topic or topic in path)
    )
    return nodes


def get_topics(topic: str = None) -> List[Sensor]:
    nodes = get_topic_inodes(topic)
    return (Sensor(node, path=path) for node, path in nodes)


@autocompeletion.complete
class uORBDump(gdb.Command):
    """Dump uORB topics"""

    formatter = "{:<20} {:<24} {:<6} {:<6} {:<6} {:<6} {:<12} {:<12} {:<20}"
    header = (
        "Address",
        "Topic",
        "Subs",
        "Ads",
        "esize",
        "nbuf",
        "latency",
        "interval",
        "Circbuf",
    )

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "--topic",
            type=str,
            help="The topic name to dump, e.g. 'sensor_accel'",
            default=None,
        )
        return parser

    def __init__(self):
        super().__init__("uorb", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        try:
            args = self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            return

        print(self.formatter.format(*self.header))
        for topic in get_topics(topic=args.topic):
            print(
                self.formatter.format(
                    hex(topic),
                    topic.topicname,
                    topic.nsubscribers,
                    topic.nadvertisers,
                    topic.state.esize,
                    topic.state.nbuffer,
                    topic.state.min_latency,
                    topic.state.min_interval,
                    hex(topic.buffer.address),
                )
            )

    def diagnose(self, *args, **kwargs):
        return {
            "title": "UORB Topics",
            "summary": "All active uORB topics",
            "command": "uorb",
            "result": "info",
            "category": utils.DiagnoseCategory.sensor,
            "message": gdb.execute("uorb", to_string=True),
        }



############################################################################
# tools/pynuttx/nxgdb/fs/lrofs.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import nxgdb.utils as utils
from nxgdb.fs.inode import Inode


def dump_lrofs_mpt(mpt):
    sector_size = mpt.lm_hwsectorsize
    sector_num = mpt.lm_hwnsectors
    volume_size = mpt.lm_volsize
    xip_base = mpt.lm_xipbase
    buffer_base = mpt.lm_buffer
    print(
        f" HW sector size: {sector_size} HW sector number: {sector_num}\n"
        f" Volume size: {volume_size} XIP_addr: {hex(xip_base)}"
        f" Buffer_addr: {hex(buffer_base)}"
    )


def dump_lrofs_files(node, level=1, prefix="", maxlevel=4096):
    if level > maxlevel:
        return

    if node.ln_count > 0:
        initial_indent = prefix + "├── "
        newprefix = prefix + "│   "
        dirfix = "/"
    else:
        initial_indent = prefix + "└── "
        newprefix = prefix + "    "
        dirfix = ""

    name = node.ln_name.string(length=node.ln_namesize)
    print(
        f"{initial_indent}{name}{dirfix} offset:{node.ln_offset} next:{node.ln_next}"
        f" size:{node.ln_size} child_count:{node.ln_count}"
    )

    for child in utils.ArrayIterator(node.ln_child, node.ln_count):
        dump_lrofs_files(child, level + 1, newprefix, maxlevel)


def dump_lrofs_cache(node: Inode, path):
    mpt = node.i_private.cast(utils.lookup_type("struct lrofs_mountpt_s").pointer())
    root = mpt.lm_root.cast(utils.lookup_type("struct lrofs_nodeinfo_s").pointer())
    print(f"Lrofs {path} mount point information: {hex(mpt)}")
    dump_lrofs_mpt(mpt)
    dump_lrofs_files(root)



############################################################################
# tools/pynuttx/gdbrpc/examples/breakpoint.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
import queue

import gdbrpc

# Set a breakpoint and notify client when hit
# Note: temporary breakpoint is used here
# Note: backtrace is printed when breakpoint is hit


class SetBreakpoint(gdbrpc.Request):
    def __init__(self, location: str):
        super().__init__()
        self.location = location

    def __call__(self, q: queue.Queue):
        import gdb

        class Bp(gdb.Breakpoint):
            import queue

            def __init__(self, spec: str, q: queue.Queue):
                # Note that we use temporary breakpoint, so that it is automatically removed after being hit
                super().__init__(spec, internal=False, temporary=True)

                # The q is used to notify GDBRPC that the __call__ is finally finished.
                self._q = q

            def stop(self) -> bool:
                # Since we are in GDB's context, we can use gdb module directly
                bt = gdb.execute("bt", to_string=True)

                # Put any message to the queue to notify breakpoint hit
                self._q.put(bt)
                return True

        bp = Bp(self.location, q)
        assert bp.is_valid()
        return self.location


class BreakpointHit(gdbrpc.PostRequest):
    def __init__(self):
        super().__init__()

    def __call__(self, result):
        print(f"Breakpoint hit!:\n{result}")


if __name__ == "__main__":

    client = gdbrpc.Client(
        log_level=logging.ERROR
    )  # Use default localhost:20819, change if needed

    assert client.connect()

    bp = SetBreakpoint("up_idle")
    notifier = BreakpointHit()
    print(client.call(bp, post_request=notifier, timeout=10))

    # Now you can do other things, or wait for breakpoint hit notification
    print("Waiting for breakpoint to be hit...")
    notifier.finish.wait()
    print("Breakpoint handling completed.")
    client.disconnect()



############################################################################
# tools/pynuttx/rpmsg.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse

import gdb

from . import autocompeletion, utils
from .lists import NxList

RPMSG_HOST = 0
RPMSG_DEVICE = 1

CONFIG_RPMSG = utils.lookup_type("struct rpmsg_s") is not None


@autocompeletion.complete
class RPMsgDump(gdb.Command):
    """Dump rpmsg service"""

    CALLBACK_HEADER = ["rpmsg_cb_s at", "ns_match", "ns_bind"]
    ENDPOINT_HEADER = [
        "name",
        "addr",
        "dest",
        "cb",
        "ns_bound_cb",
        "ns_unbind_cb",
        "endpoint",
    ]
    VQ_HEADER = [
        "vq",
        "name",
        "size",
        "free",
        "queued",
        "desc_head_idx",
        "available_idx",
        "avail.idx",
        "avail.flags",
        "used_cons_idx",
        "used.idx",
        "used.flags",
    ]

    CALLBACK_FORAMTTER = "{:<20} {:<40} {:<40}"
    ENDPOINT_FORMATTER = "{:<24} {:<5} {:<11} {:<40} {:<40} {:<40} {:<20}"
    VQ_FORMATTER = (
        "{:<12} {:<8} {:<5} {:<5} {:<7} {:<14} {:<14} {:<10} {:<12} {:<14} {:<9} {:<11}"
    )

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-f",
            "--full",
            action="store_true",
            help="Dump the rpmsg information fullly",
        )
        return parser

    def __init__(self):
        if CONFIG_RPMSG:
            super(RPMsgDump, self).__init__("rpmsgdump", gdb.COMMAND_USER)
            self.parser = self.get_argparser()

    def parse_args(self, arg):
        try:
            return self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            return

    def is_rpmsg_transport(self, rdev, transport):
        if transport == "virtio":
            tx_payload = "rpmsg_virtio_get_tx_payload_buffer"
        elif transport == "port":
            tx_payload = "rpmsg_port_get_tx_payload_buffer"
        else:
            return False

        symbol = gdb.lookup_symbol(tx_payload)
        if symbol[0] is None:
            return False

        real_addr = symbol[0].value().address
        ops_addr = int(utils.Value(rdev["ops"]["get_tx_payload_buffer"])) & ~1
        return real_addr == ops_addr

    def dump_virtqueue(self, vq):
        vr = vq["vq_ring"]
        gdb.write(
            self.VQ_FORMATTER.format(
                f"{vq}",
                f"{vq['vq_name'].string()}",
                f"{vq['vq_nentries']}",
                f"{vq['vq_free_cnt']}",
                f"{vq['vq_queued_cnt']}",
                f"{vq['vq_desc_head_idx']}",
                f"{vq['vq_available_idx']}",
                f"{vr['avail']['idx']}",
                f"{vr['avail']['flags']}",
                f"{vq['vq_used_cons_idx']}",
                f"{vr['used']['idx']}",
                f"{vr['used']['flags']}",
            )
            + "\n"
        )

    def rpmsg_virtio_buffer_nused(self, rvdev, rx):
        vq = rvdev["rvq"] if rx else rvdev["svq"]
        is_host = rvdev["vdev"]["role"] == RPMSG_HOST

        nused = vq["vq_ring"]["avail"]["idx"] - vq["vq_ring"]["used"]["idx"]
        if is_host ^ rx:
            return nused
        else:
            return vq["vq_nentries"] - nused

    def dump_rpmsg_virtio_buffer(self, rvdev, rx):
        vq = rvdev["rvq"] if rx else rvdev["svq"]
        nused = self.rpmsg_virtio_buffer_nused(rvdev, rx)

        gdb.write(
            f"    {'RX' if rx else 'Tx'} buffer, total {vq['vq_nentries']}, pending {nused}\n"
        )

        for i in range(nused):
            if (rvdev["vdev"]["role"] == RPMSG_HOST) ^ rx:
                desc_idx = (vq["vq_ring"]["used"]["idx"] + i) & (vq["vq_nentries"] - 1)
                desc_idx = vq["vq_ring"]["avail"]["ring"][desc_idx]
            else:
                desc_idx = (vq["vq_ring"]["avail"]["idx"] + i) & (vq["vq_nentries"] - 1)
                desc_idx = vq["vq_ring"]["used"]["ring"][desc_idx]["id"]

            addr = vq["vq_ring"]["desc"][desc_idx]["addr"]
            if addr:
                hdr = addr.cast(utils.lookup_type("struct rpmsg_hdr").pointer())
                ept = self.rpmsg_get_ept_from_addr(
                    rvdev["rdev"], hdr["dst"] if rx else hdr["src"]
                )

                if ept:
                    gdb.write(
                        f"      {'RX' if rx else 'TX'} buffer:{addr} held by {ept['name']}\n"
                    )

    def dump_rpmsg_virtio(self, rdev):
        if not self.is_rpmsg_transport(rdev, "virtio"):
            return

        rvdev = rdev.cast(utils.lookup_type("struct rpmsg_virtio_device").pointer())

        gdb.write("Rpmsg VirtIO Transport:\n")
        gdb.write(
            f"rvdev:{rvdev} h2r_buf_size:{rvdev['config']['h2r_buf_size']}  "
            f"r2h_buf_size:{rvdev['config']['r2h_buf_size']}\n"
        )
        for vbuff in NxList(rvdev["reclaimer"], "struct vbuff_reclaimer_t", "node"):
            gdb.write(f"rvdev reclaimer vbuff:{vbuff} idx:{vbuff['idx']}")

        gdb.write("Rpmsg Virtqueues:\n")
        self.print_headers(self.VQ_HEADER, self.VQ_FORMATTER)
        try:
            self.dump_virtqueue(rvdev["svq"])
            self.dump_virtqueue(rvdev["rvq"])
            gdb.write("  rpmsg buffer list:\n")
            self.dump_rpmsg_virtio_buffer(rvdev, True)
            self.dump_rpmsg_virtio_buffer(rvdev, False)

        except gdb.error as e:
            gdb.write(f"Error when dump virtqueues: {e}\n")

        gdb.write("\n")

    def rpmsg_get_ept_from_addr(self, rdev, addr):
        endpoints = rdev["endpoints"]
        for endpoint in NxList(endpoints, "struct rpmsg_endpoint", "node"):
            ept_addr = endpoint["addr"]
            if ept_addr == addr:
                return endpoint
        return None

    def rpmsg_port_node_to_buf(self, queue, node):
        node_offset = (int(node) - int(queue["node"])) / int(
            utils.sizeof("struct list_node")
        )
        buf_addr = queue["buf"] + (int(node_offset) * queue["len"])
        return buf_addr

    def dump_rpmsg_port_buffer(self, rdev, queue, label):
        buffer_list = []
        head = queue["ready"]["head"]

        for node in NxList(head):
            hdr = self.rpmsg_port_node_to_buf(queue, node)
            hdr = hdr.cast(utils.lookup_type("struct rpmsg_port_header_s").pointer())
            if not hdr or not hdr["buf"]:
                continue

            rphdr = hdr["buf"].cast(utils.lookup_type("struct rpmsg_hdr").pointer())
            ept_addr = rphdr["dst"] if label == "RX" else rphdr["src"]
            ept = self.rpmsg_get_ept_from_addr(rdev, ept_addr)
            if ept:
                ept_name = ept["name"].string().split("\0", 1)[0]
                buffer_list.append(f"{label} buffer:{rphdr} held by {ept_name}\n")

        return buffer_list

    def dump_rpmsg_port(self, rdev):
        if not self.is_rpmsg_transport(rdev, "port"):
            return

        port = utils.container_of(rdev, "struct rpmsg_port_s", "rdev")

        gdb.write(f"rxq nused:{port['rxq']['ready']['num']}\n")
        gdb.write(f"rxq navail:{port['rxq']['free']['num']}\n")
        gdb.write(f"txq nused:{port['txq']['ready']['num']}\n")
        gdb.write(f"txq navail:{port['txq']['free']['num']}\n")

        rx_buffers = self.dump_rpmsg_port_buffer(rdev, port["rxq"], "RX")
        tx_buffers = self.dump_rpmsg_port_buffer(rdev, port["txq"], "TX")
        for buffer in rx_buffers + tx_buffers:
            gdb.write(buffer + "\n")

    def print_headers(self, headers, formatter):
        gdb.write(formatter.format(*headers) + "\n")
        gdb.write(formatter.format(*["-" * len(header) for header in headers]) + "\n")

    def dump_rdev_epts(self, endpoints_head):
        gdb.write("Rpmsg Device Endpoints:\n")
        gdb.write(f"endpoints list:{endpoints_head}\n")
        self.print_headers(self.ENDPOINT_HEADER, self.ENDPOINT_FORMATTER)

        output = []
        for endpoint in NxList(endpoints_head, "struct rpmsg_endpoint", "node"):
            output.append(
                self.ENDPOINT_FORMATTER.format(
                    f"{endpoint['name'].string()}",
                    f"{endpoint['addr']}",
                    f"{endpoint['dest_addr']}",
                    f"{endpoint['cb']}",
                    f"{endpoint['ns_bound_cb']}",
                    f"{endpoint['ns_unbind_cb']}",
                    f"{endpoint}",
                )
            )

        gdb.write("\n".join(output) + "\n")

    def dump_rdev_bitmap(self, rdev):
        bitmap_values = [hex(bit) for bit in utils.ArrayIterator(rdev["bitmap"])]

        gdb.write(
            f"bitmap: {' '.join(bitmap_values):<20} bitnext: {hex(rdev['bitnext'])}\n"
        )

    def dump_rdev(self, rdev):
        gdb.write("Rpmsg Device:\n")
        gdb.write(f"rdev:{rdev}\n")
        self.dump_rdev_bitmap(rdev)
        self.dump_rdev_epts(rdev["endpoints"])

    def dump_rpmsg_cb(self):
        gdb.write("Rpmsg Callback:\n")
        self.print_headers(self.CALLBACK_HEADER, self.CALLBACK_FORAMTTER)

        output = []
        for cb in NxList(
            utils.parse_and_eval("g_rpmsg_cb"), "struct rpmsg_cb_s", "node"
        ):
            output.append(
                self.CALLBACK_FORAMTTER.format(
                    str(cb), str(cb["ns_match"]), str(cb["ns_bind"])
                )
            )
        gdb.write("\n".join(output) + "\n")

    def dump_rpmsg(self, args):
        for rpmsg in NxList(utils.parse_and_eval("g_rpmsg"), "struct rpmsg_s", "node"):
            rdev = utils.Value(int(rpmsg) + utils.sizeof("struct rpmsg_s"))
            rdev = rdev.cast(utils.lookup_type("struct rpmsg_device").pointer())
            gdb.write(
                f"Rpmsg:\n"
                f"rpmsg:{rpmsg} localcpu:{rpmsg['local_cpuname']} remotecpu:{rpmsg['cpuname']}\n"
            )
            if args.full:
                self.dump_rdev(rdev)
            self.dump_rpmsg_virtio(rdev)
            self.dump_rpmsg_port(rdev)

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        if not (args := self.parse_args(args)):
            return

        if args.full:
            self.dump_rpmsg_cb()
        self.dump_rpmsg(args)

    def diagnose(self, *args, **kwargs):
        return {
            "title": "RPMSG Report",
            "summary": "RPMSG report",
            "command": "rpmsgdump",
            "result": "info",
            "category": utils.DiagnoseCategory.rpc,
            "message": gdb.execute("rpmsgdump", to_string=True),
        }



############################################################################
# tools/pynuttx/tests/test_mock_stack.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import unittest
from unittest.mock import MagicMock, patch

import gdb
from nxgdb.stack import Stack, fetch_stacks


class TestStack(unittest.TestCase):
    """
    This test basically creates some mock objects to verify the
    correctness of the Stack Class
    """

    def test_stack_init_badsize(self):
        with self.assertWarns(RuntimeWarning):
            name = "test_thread"
            entry = hex(0xABCD)
            base = 0x2000
            alloc = 0x1000
            size = 0x0
            cursp = 0x1500
            align = 4
            Stack(name, entry, base, alloc, size, cursp, align)

    def test_stack_init_bad_allocaddr(self):
        with self.assertWarns(RuntimeWarning):
            name = "test_thread"
            entry = hex(0xABCD)
            alloc = 0x1000
            base = alloc - 1
            size = 0x1000
            cursp = 0x1500
            align = 4
            Stack(name, entry, base, alloc, size, cursp, align)

    def test_stack_init_bad_sp(self):
        with self.assertWarns(RuntimeWarning):
            name = "test_thread"
            entry = hex(0xABCD)
            base = 0x2000
            alloc = 0x1000
            size = 0x1000
            cursp = base - 1
            align = 4
            Stack(name, entry, base, alloc, size, cursp, align)

        with self.assertWarns(RuntimeWarning):
            name = "test_thread"
            entry = hex(0xABCD)
            base = 0x2000
            alloc = 0x1000
            size = 0x1000
            cursp = base + size + 1
            align = 4
            Stack(name, entry, base, alloc, size, cursp, align)

    @patch("gdb.write")
    @patch("gdb.lookup_type")
    @patch("gdb.Value")
    def test_stack_cur_usage_no_overflow(
        self, mock_gdb_value, mock_gdb_lookup_type, mock_gdb_write
    ):
        name = "test_thread"
        entry = hex(0xABCD)
        base = 0x1000
        alloc = 0x900
        size = 0x1000
        cursp = 0x1500
        align = 4
        stack = Stack(name, entry, base, alloc, size, cursp, align)

        mock_gdb_value.return_value.cast.return_value = [0xC0FEBABE] * (size // align)
        mock_gdb_lookup_type.return_value.pointer.return_value = MagicMock()

        self.assertEqual(stack.cur_usage(), base + size - cursp)
        mock_gdb_write.assert_not_called()

    @patch("gdb.write")
    @patch("gdb.lookup_type")
    @patch("gdb.Value")
    def test_stack_cur_usage_with_overflow(
        self, mock_gdb_value, mock_gdb_lookup_type, mock_gdb_write
    ):
        name = "test_thread"
        entry = hex(0xABCD)
        base = 0x1000
        alloc = 0x900
        size = 0x1000
        cursp = 0x1500
        align = 4
        stack = Stack(name, entry, base, alloc, size, cursp, align)

        mock_gdb_value.return_value.cast.return_value = [0xC0FEBABE] * (size // align)
        mock_gdb_lookup_type.return_value.pointer.return_value = MagicMock()

        # After constructing the stack, we modify the stack sp to mock an
        # overflow behaviour
        stack._cur_sp = base - align

        with self.assertRaises(gdb.GdbError) as context:
            stack.cur_usage()

        self.assertTrue("pls check your stack size!" in str(context.exception))

    @patch("nxgdb.utils.get_symbol_value")
    def test_stack_check_max_usage_no_color(self, mock_utils_get_symbol_value):
        name = "test_thread"
        entry = hex(0xABCD)
        base = 0x1000
        alloc = 0x900
        size = 0x1000
        cursp = 0x1500
        align = 4
        stack = Stack(name, entry, base, alloc, size, cursp, align)

        mock_utils_get_symbol_value.return_value = None

        self.assertEqual(stack.max_usage(), 0)


class TestFetchStacks(unittest.TestCase):
    @patch("nxgdb.utils.is_target_arch")
    @patch("nxgdb.utils.in_interrupt_context")
    @patch("nxgdb.utils.get_register_byname")
    @patch("nxgdb.utils.get_tcbs")
    @patch("nxgdb.utils.get_task_name")
    def test_fetch_stacks(self, *args):
        (
            mock_get_task_name,
            mock_get_tcbs,
            mock_get_register_byname,
            mock_in_interrupt_context,
            mock_is_target_arch,
        ) = args

        mock_get_tcbs.return_value = [
            {
                "task_state": 3,
                "pid": 123,
                "name": "test",
                "entry": {"pthread": 0x1000},
                "stack_base_ptr": 0x2000,
                "stack_alloc_ptr": 0x1000,
                "adj_stack_size": 0x4000,
            }
        ]
        mock_get_task_name.return_value = "test"
        mock_is_target_arch.return_value = True
        mock_in_interrupt_context.return_value = False
        mock_get_register_byname.return_value = 0x5000
        gdb.parse_and_eval = MagicMock(
            side_effect=(
                lambda x: (
                    mock_get_tcbs.return_value
                    if x == "g_pidhash"
                    else (1 if x == "g_npidhash" else 3)
                )
            )
        )

        stacks = fetch_stacks()

        self.assertNotEqual(stacks, dict())
        self.assertEqual(len(stacks), 1)
        self.assertIn(123, stacks)
        self.assertIsNotNone(stacks[123])
        stack = stacks[123]
        self.assertEqual(str(stack._thread_name), "test")
        self.assertEqual(stack._thread_entry, hex(0x1000))
        self.assertEqual(stack._stack_base, 0x2000)
        self.assertEqual(stack._stack_alloc, 0x1000)
        self.assertEqual(stack._stack_size, 0x4000)
        self.assertEqual(stack._cur_sp, 0x5000)
        self.assertEqual(stack._align, 4)

        # If we get an error while reading the register, we expect an
        # empry stack object
        mock_get_register_byname.side_effect = gdb.GdbError
        with self.assertRaises(gdb.GdbError):
            stacks = fetch_stacks()
            self.assertEqual(stacks, dict())



############################################################################
# tools/pynuttx/nxgdb/fs/yaffs.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import nxgdb.utils as utils
from nxgdb.fs.inode import Inode
from nxgdb.lists import NxList

YAFFS_FILETYPE_MAP = {
    0: "yaffs_UNKNOWN_var",
    1: "yaffs_file_var",
    2: "yaffs_symlink_var",
    3: "yaffs_dir_var",
    4: "yaffs_hardlink_var",
    5: "yaffs_SPECIAL_var",
}


def dump_yaffs_files(node, level=1, prefix="", maxlevel=4096):
    if level > maxlevel:
        return

    file_type = YAFFS_FILETYPE_MAP[int(node.variant_type)]
    if file_type == "yaffs_dir_var":
        initial_indent = prefix + "├── "
        newprefix = prefix + "│   "
        dirfix = "/"
        moreinfo = ""
    elif file_type == "yaffs_file_var":
        initial_indent = prefix + "└── "
        newprefix = prefix + "    "
        dirfix = ""
        file = node.variant.cast(utils.lookup_type("struct yaffs_file_var"))
        file_size = file.file_size
        stored = file.stored_size
        moreinfo = f" file_size:{file_size},stored:{stored}"
    else:
        initial_indent = prefix + "└── "
        newprefix = prefix + "    "
        dirfix = ""
        moreinfo = ""

    obj_id = node.obj_id
    short_name = node.short_name.string(length=16).rstrip("\0")
    if short_name == "":
        short_name = "(null)"
    hdr_chunk = node.hdr_chunk
    n_data_chunks = node.n_data_chunks
    serial = str(node.serial).split(" ")[0]
    baseinfo = {
        f"name:{short_name}{dirfix},obj_id:{obj_id},hdr_chunk:{hdr_chunk},"
        f"n_data_chunks:{n_data_chunks},serial:{serial}{moreinfo}"
    }

    print(f"{initial_indent}{file_type.rsplit('_', 2)[1]} {baseinfo}")
    if file_type == "yaffs_dir_var":
        head = node.variant.cast(utils.lookup_type("struct yaffs_dir_var"))
        for siblings in NxList(head.children, "struct yaffs_obj", "siblings"):
            dump_yaffs_files(siblings, level + 1, newprefix, maxlevel)


def dump_yaffs_mpt(mpt):
    geo = mpt.geo.cast(utils.lookup_type("struct mtd_geometry_s"))
    dev = mpt.dev.cast(utils.lookup_type("struct yaffs_dev").pointer())
    blksize = geo.blocksize
    erasesize = geo.erasesize
    erasecount = geo.neraseblocks
    print(f" Block size:{blksize} Erase size:{erasesize} Erase count:{erasecount}")
    print(f"Dump yaffs_dev: {hex(dev)}")
    print(dev.dereference().format_string(pretty_structs=True, styling=True))


def dump_yaffs_cache(node: Inode, path):
    mpt = node.i_private.cast(utils.lookup_type("struct yaffs_mountpt_s").pointer())
    dev = mpt.dev.cast(utils.lookup_type("struct yaffs_dev").pointer())
    root = dev.root_dir.cast(utils.lookup_type("struct yaffs_obj").pointer())
    print(f"Yaffs {path} mount point information: {hex(mpt)}")
    dump_yaffs_mpt(mpt)
    print(f"Yaffs {path} file tree information: {hex(root)}")
    dump_yaffs_files(root)
